@@ -1,7 +1,10 @@
-from greentest import TestCase, main
+import greentest
+from greentest import TestCase, main, GenericGetTestCase
 import gevent
-from gevent import util, core
+from gevent.hub import get_hub, LoopExit
+from gevent import util
 from gevent import queue
+from gevent.queue import Empty, Full
 from gevent.event import AsyncResult
 
 
@@ -11,23 +14,43 @@ class TestQueue(TestCase):
         self.switch_expected = False
         q = queue.Queue()
         q.put('hi')
-        self.assertEquals(q.get(), 'hi')
+        self.assertEqual(q.peek(), 'hi')
+        self.assertEqual(q.get(), 'hi')
+
+    def test_peek_empty(self):
+        q = queue.Queue()
+        # No putters waiting, in the main loop: LoopExit
+        self.assertRaises(LoopExit, q.peek)
+
+        def waiter(q):
+            self.assertRaises(Empty, q.peek, timeout=0.01)
+        g = gevent.spawn(waiter, q)
+        gevent.sleep(0.1)
+        g.join()
+
+    def test_peek_multi_greenlet(self):
+        q = queue.Queue()
+        g = gevent.spawn(q.peek)
+        g.start()
+        gevent.sleep(0)
+        q.put(1)
+        g.join()
+        self.assertTrue(g.exception is None)
+        self.assertEqual(q.peek(), 1)
 
     def test_send_last(self):
         q = queue.Queue()
+
         def waiter(q):
-            timer = gevent.Timeout.start_new(0.1)
-            try:
-                self.assertEquals(q.get(), 'hi2')
-            finally:
-                timer.cancel()
+            with gevent.Timeout(0.1 if not greentest.RUNNING_ON_APPVEYOR else 0.5):
+                self.assertEqual(q.get(), 'hi2')
             return "OK"
 
         p = gevent.spawn(waiter, q)
         gevent.sleep(0.01)
         q.put('hi2')
         gevent.sleep(0.01)
-        assert p.get(timeout=0)=="OK"
+        assert p.get(timeout=0) == "OK"
 
     def test_max_size(self):
         q = queue.Queue(2)
@@ -44,16 +67,17 @@ class TestQueue(TestCase):
 
         p = gevent.spawn(putter, q)
         gevent.sleep(0)
-        self.assertEquals(results, ['a', 'b'])
-        self.assertEquals(q.get(), 'a')
+        self.assertEqual(results, ['a', 'b'])
+        self.assertEqual(q.get(), 'a')
         gevent.sleep(0)
-        self.assertEquals(results, ['a', 'b', 'c'])
-        self.assertEquals(q.get(), 'b')
-        self.assertEquals(q.get(), 'c')
-        assert p.get(timeout=0)=="OK"
+        self.assertEqual(results, ['a', 'b', 'c'])
+        self.assertEqual(q.get(), 'b')
+        self.assertEqual(q.get(), 'c')
+        assert p.get(timeout=0) == "OK"
 
     def test_zero_max_size(self):
-        q = queue.Queue(0)
+        q = queue.Channel()
+
         def sender(evt, q):
             q.put('hi')
             evt.set('done')
@@ -67,15 +91,12 @@ class TestQueue(TestCase):
 
         p1 = gevent.spawn(sender, e1, q)
         gevent.sleep(0.001)
-        self.assert_(not e1.ready())
+        self.assertTrue(not e1.ready())
         p2 = gevent.spawn(receiver, e2, q)
-        self.assertEquals(e2.get(),'hi')
-        self.assertEquals(e1.get(),'done')
-        timeout = gevent.Timeout.start_new(0)
-        try:
+        self.assertEqual(e2.get(), 'hi')
+        self.assertEqual(e1.get(), 'done')
+        with gevent.Timeout(0):
             gevent.joinall([p1, p2])
-        finally:
-            timeout.cancel()
 
     def test_multiple_waiters(self):
         # tests that multiple waiters get their results back
@@ -87,46 +108,44 @@ class TestQueue(TestCase):
         sendings = ['1', '2', '3', '4']
         evts = [AsyncResult() for x in sendings]
         for i, x in enumerate(sendings):
-            gevent.spawn(waiter, q, evts[i]) # use waitall for them
+            gevent.spawn(waiter, q, evts[i])  # XXX use waitall for them
 
-        gevent.sleep(0.01) # get 'em all waiting
+        gevent.sleep(0.01)  # get 'em all waiting
 
         results = set()
+
         def collect_pending_results():
-            for i, e in enumerate(evts):
-                timer = gevent.Timeout.start_new(0.001)
-                try:
+            for e in evts:
+                with gevent.Timeout(0.001, False):
                     x = e.get()
                     results.add(x)
-                    timer.cancel()
-                except gevent.Timeout:
-                    pass  # no pending result at that event
             return len(results)
+
         q.put(sendings[0])
-        self.assertEquals(collect_pending_results(), 1)
+        self.assertEqual(collect_pending_results(), 1)
         q.put(sendings[1])
-        self.assertEquals(collect_pending_results(), 2)
+        self.assertEqual(collect_pending_results(), 2)
         q.put(sendings[2])
         q.put(sendings[3])
-        self.assertEquals(collect_pending_results(), 4)
+        self.assertEqual(collect_pending_results(), 4)
 
     def test_waiters_that_cancel(self):
         q = queue.Queue()
 
         def do_receive(q, evt):
-            gevent.Timeout.start_new(0, RuntimeError())
-            try:
-                result = q.get()
-                evt.set(result)
-            except RuntimeError:
-                evt.set('timed out')
+            with gevent.Timeout(0, RuntimeError()):
+                try:
+                    result = q.get()
+                    evt.set(result)
+                except RuntimeError:
+                    evt.set('timed out')
 
         evt = AsyncResult()
         gevent.spawn(do_receive, q, evt)
-        self.assertEquals(evt.get(), 'timed out')
+        self.assertEqual(evt.get(), 'timed out')
 
         q.put('hi')
-        self.assertEquals(q.get(), 'hi')
+        self.assertEqual(q.get(), 'hi')
 
     def test_senders_that_die(self):
         q = queue.Queue()
@@ -135,58 +154,56 @@ class TestQueue(TestCase):
             q.put('sent')
 
         gevent.spawn(do_send, q)
-        self.assertEquals(q.get(), 'sent')
+        self.assertEqual(q.get(), 'sent')
 
     def test_two_waiters_one_dies(self):
+
         def waiter(q, evt):
             evt.set(q.get())
+
         def do_receive(q, evt):
-            timeout = gevent.Timeout.start_new(0, RuntimeError())
-            try:
+            with gevent.Timeout(0, RuntimeError()):
                 try:
                     result = q.get()
                     evt.set(result)
                 except RuntimeError:
                     evt.set('timed out')
-            finally:
-                timeout.cancel()
 
         q = queue.Queue()
         dying_evt = AsyncResult()
         waiting_evt = AsyncResult()
         gevent.spawn(do_receive, q, dying_evt)
         gevent.spawn(waiter, q, waiting_evt)
-        gevent.sleep(0)
+        gevent.sleep(0.1)
         q.put('hi')
-        self.assertEquals(dying_evt.get(), 'timed out')
-        self.assertEquals(waiting_evt.get(), 'hi')
+        self.assertEqual(dying_evt.get(), 'timed out')
+        self.assertEqual(waiting_evt.get(), 'hi')
 
     def test_two_bogus_waiters(self):
         def do_receive(q, evt):
-            gevent.Timeout.start_new(0, RuntimeError())
-            try:
-                result = q.get()
-                evt.set(result)
-            except RuntimeError:
-                evt.set('timed out')
-            # XXX finally = timeout
+            with gevent.Timeout(0, RuntimeError()):
+                try:
+                    result = q.get()
+                    evt.set(result)
+                except RuntimeError:
+                    evt.set('timed out')
 
         q = queue.Queue()
         e1 = AsyncResult()
         e2 = AsyncResult()
         gevent.spawn(do_receive, q, e1)
         gevent.spawn(do_receive, q, e2)
-        gevent.sleep(0)
+        gevent.sleep(0.1)
         q.put('sent')
-        self.assertEquals(e1.get(), 'timed out')
-        self.assertEquals(e2.get(), 'timed out')
-        self.assertEquals(q.get(), 'sent')
+        self.assertEqual(e1.get(), 'timed out')
+        self.assertEqual(e2.get(), 'timed out')
+        self.assertEqual(q.get(), 'sent')
 
 
 class TestChannel(TestCase):
 
     def test_send(self):
-        channel = queue.Queue(0)
+        channel = queue.Channel()
 
         events = []
 
@@ -206,7 +223,7 @@ class TestChannel(TestCase):
         g.get()
 
     def test_wait(self):
-        channel = queue.Queue(0)
+        channel = queue.Channel()
         events = []
 
         def another_greenlet():
@@ -227,48 +244,93 @@ class TestChannel(TestCase):
         self.assertEqual(['waiting', 'sending hello', 'hello', 'sending world', 'world', 'sent world'], events)
         g.get()
 
+    def test_task_done(self):
+        channel = queue.JoinableQueue(0)
+        X = object()
+        gevent.spawn(channel.put, X)
+        result = channel.get()
+        assert result is X, (result, X)
+        assert channel.unfinished_tasks == 1, channel.unfinished_tasks
+        channel.task_done()
+        assert channel.unfinished_tasks == 0, channel.unfinished_tasks
+
+    def test_iterable(self):
+        channel = queue.Channel()
+        gevent.spawn(channel.put, StopIteration)
+        r = list(channel)
+        self.assertEqual(r, [])
+
 
 class TestNoWait(TestCase):
 
     def test_put_nowait_simple(self):
         result = []
         q = queue.Queue(1)
+
         def store_result(func, *args):
             result.append(func(*args))
-        core.active_event(store_result, util.wrap_errors(Exception, q.put_nowait), 2)
-        core.active_event(store_result, util.wrap_errors(Exception, q.put_nowait), 3)
+
+        run_callback = get_hub().loop.run_callback
+
+        run_callback(store_result, util.wrap_errors(Full, q.put_nowait), 2)
+        run_callback(store_result, util.wrap_errors(Full, q.put_nowait), 3)
         gevent.sleep(0)
-        assert len(result)==2, result
-        assert result[0]==None, result
+        assert len(result) == 2, result
+        assert result[0] is None, result
         assert isinstance(result[1], queue.Full), result
 
     def test_get_nowait_simple(self):
         result = []
         q = queue.Queue(1)
         q.put(4)
+
         def store_result(func, *args):
             result.append(func(*args))
-        core.active_event(store_result, util.wrap_errors(Exception, q.get_nowait))
-        core.active_event(store_result, util.wrap_errors(Exception, q.get_nowait))
+
+        run_callback = get_hub().loop.run_callback
+
+        run_callback(store_result, util.wrap_errors(Empty, q.get_nowait))
+        run_callback(store_result, util.wrap_errors(Empty, q.get_nowait))
         gevent.sleep(0)
-        assert len(result)==2, result
-        assert result[0]==4, result
+        assert len(result) == 2, result
+        assert result[0] == 4, result
         assert isinstance(result[1], queue.Empty), result
 
     # get_nowait must work from the mainloop
     def test_get_nowait_unlock(self):
         result = []
-        q = queue.Queue(0)
+        q = queue.Queue(1)
         p = gevent.spawn(q.put, 5)
+
         def store_result(func, *args):
             result.append(func(*args))
+
         assert q.empty(), q
+        gevent.sleep(0)
         assert q.full(), q
+        get_hub().loop.run_callback(store_result, q.get_nowait)
         gevent.sleep(0)
+        assert q.empty(), q
+        assert result == [5], result
+        assert p.ready(), p
+        assert p.dead, p
+        assert q.empty(), q
+
+    def test_get_nowait_unlock_channel(self):
+        result = []
+        q = queue.Channel()
+        p = gevent.spawn(q.put, 5)
+
+        def store_result(func, *args):
+            result.append(func(*args))
+
         assert q.empty(), q
         assert q.full(), q
-        core.active_event(store_result, util.wrap_errors(Exception, q.get_nowait))
-        gevent.sleep(0)
+        gevent.sleep(0.001)
+        assert q.empty(), q
+        assert q.full(), q
+        get_hub().loop.run_callback(store_result, q.get_nowait)
+        gevent.sleep(0.001)
         assert q.empty(), q
         assert q.full(), q
         assert result == [5], result
@@ -279,23 +341,77 @@ class TestNoWait(TestCase):
     # put_nowait must work from the mainloop
     def test_put_nowait_unlock(self):
         result = []
-        q = queue.Queue(0)
+        q = queue.Queue()
         p = gevent.spawn(q.get)
+
         def store_result(func, *args):
             result.append(func(*args))
+
         assert q.empty(), q
-        assert q.full(), q
-        gevent.sleep(0)
+        assert not q.full(), q
+        gevent.sleep(0.001)
         assert q.empty(), q
-        assert q.full(), q
-        core.active_event(store_result, util.wrap_errors(Exception, q.put_nowait), 10)
+        assert not q.full(), q
+        get_hub().loop.run_callback(store_result, q.put_nowait, 10)
         assert not p.ready(), p
-        gevent.sleep(0)
+        gevent.sleep(0.001)
         assert result == [None], result
         assert p.ready(), p
-        assert q.full(), q
+        assert not q.full(), q
         assert q.empty(), q
 
 
-if __name__=='__main__':
+class TestJoinEmpty(TestCase):
+
+    def test_issue_45(self):
+        """Test that join() exits immediatelly if not jobs were put into the queue"""
+        self.switch_expected = False
+        q = queue.JoinableQueue()
+        q.join()
+
+
+def make_get_interrupt(queue_type):
+
+    class TestGetInterrupt(GenericGetTestCase):
+
+        Timeout = Empty
+
+        def wait(self, timeout):
+            return queue_type().get(timeout=timeout)
+
+    TestGetInterrupt.__name__ += '_' + queue_type.__name__
+    return TestGetInterrupt
+
+
+for queue_type in [queue.Queue, queue.JoinableQueue, queue.LifoQueue, queue.PriorityQueue, queue.Channel]:
+    klass = make_get_interrupt(queue_type)
+    globals()[klass.__name__] = klass
+del klass, queue_type
+
+
+def make_put_interrupt(queue):
+
+    class TestPutInterrupt(GenericGetTestCase):
+
+        Timeout = Full
+
+        def wait(self, timeout):
+            while not queue.full():
+                queue.put(1)
+            return queue.put(2, timeout=timeout)
+
+    TestPutInterrupt.__name__ += '_' + queue.__class__.__name__
+    return TestPutInterrupt
+
+
+for obj in [queue.Queue(1), queue.JoinableQueue(1), queue.LifoQueue(1), queue.PriorityQueue(1), queue.Channel()]:
+    klass = make_put_interrupt(obj)
+    globals()[klass.__name__] = klass
+del klass, obj
+
+
+del GenericGetTestCase
+
+
+if __name__ == '__main__':
     main()
